@@ -5,11 +5,12 @@ Purpose: Delete GitHub Action Workflow Runs
 """
 
 import concurrent.futures
+import json
 import os
 import sys
 import threading
 import time
-from datetime import timedelta
+from datetime import datetime, timedelta
 from pathlib import Path
 
 import click
@@ -43,7 +44,7 @@ def get_auth():
     sys.exit(1)
 
 
-def get_owner_repo(repo_url):
+def get_owner_repo(repo_url: str) -> str:
     """
     Get owner/repo for pyGitHub to interact with GitHub API
 
@@ -57,7 +58,7 @@ def get_owner_repo(repo_url):
     return owner_repo
 
 
-def check_user_inputs(repo, repo_url, min_runs, max_days):
+def check_user_inputs(repo: Repository.Repository, repo_url: str, min_runs: int, max_days: int) -> bool:
     """
     Check user inputs
 
@@ -94,21 +95,28 @@ def check_user_inputs(repo, repo_url, min_runs, max_days):
     return True
 
 
-def get_core_api_rate_limit(gh):
+def get_core_api_rate_limit(gh: Github) -> tuple[int, datetime]:
     """
     Get Core API Rate Limit (rate limit endpoint itself does not consume regular API quota)
 
     Parameter(s):
     gh: github class object from get_auth()
 
-    Return: core api limit
+    Return: core api limit remaining and reset at
     """
     RateLimitOverview = gh.get_rate_limit()
-    core_limit = RateLimitOverview.resources.core
-    return core_limit
+    core = RateLimitOverview.resources.core
+    core_reset = core.reset
+    core_remaining = core.remaining
+
+    print('\n💥 Core API Rate Limit Info')
+    print(f'API rate limit remaining: {core_remaining}')
+    print(f'API rate limit reset at : {core_reset} (UTC)\n')
+
+    return (core_remaining, core_reset)
 
 
-def get_all_workflow_runs(repo):
+def get_all_workflow_runs(repo: Repository.Repository) -> pd.DataFrame:
     """
     Get all workflow runs
 
@@ -140,30 +148,12 @@ def get_all_workflow_runs(repo):
         for thread in threads:
             thread.join()
 
-        # with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor:
-        #     for runs in wf_runs:
-        #         futures = [executor.submit(
-        #             append_runs_to_list,
-        #             all_runs,
-        #             runs.workflow_id,
-        #             runs.id,
-        #             runs.created_at,
-        #             Path(runs.path).stem,
-        #             lock
-        #         )]
-        #         for future in concurrent.futures.as_completed(futures):
-        #             progress.update(overall_task, advance=1)
-        #             try:
-        #                 result = future.result()
-        #                 print(f"Task {result} Completed")
-        #             except Exception as e:
-        #                 print(f"Task Exception: {e}")
-
     df_all_runs = pd.DataFrame(all_runs)
     return df_all_runs
 
 
-def append_runs_to_list(all_runs, workflow_id, run_id, created_at, name, lock):
+def append_runs_to_list(all_runs: list, workflow_id: int, run_id: int, created_at: datetime,
+                        name: str, lock: threading.Lock):
     """
     Append workflow runs to all_runs list
 
@@ -173,8 +163,6 @@ def append_runs_to_list(all_runs, workflow_id, run_id, created_at, name, lock):
     name       : workflow run name from workflow path
     run_id     : run id from workflow
     workflow_id: workflow id from all_runs
-
-    Return: pandas dataframe that contains all workflow runs
     """
     with lock:
         all_runs.append({
@@ -185,7 +173,8 @@ def append_runs_to_list(all_runs, workflow_id, run_id, created_at, name, lock):
         })
 
 
-def break_down_df_all_runs(repo, df_all_runs):
+def break_down_df_all_runs(repo: Repository.Repository, df_all_runs: pd.DataFrame) ->\
+                           tuple[pd.DataFrame, pd.DataFrame, list]:
     """
     break down all workflow runs
 
@@ -217,7 +206,8 @@ def break_down_df_all_runs(repo, df_all_runs):
     return df_orphan_runs, df_active_runs, list_orphan_ids
 
 
-def delete_orphan_workflow_runs(repo, owner_repo, dry_run, df_orphan_runs):
+def delete_orphan_workflow_runs(repo: Repository.Repository, owner_repo: str, dry_run: bool,
+                                df_orphan_runs: pd.DataFrame) -> int:
     """
     Delete orphan workflow runs
 
@@ -227,7 +217,7 @@ def delete_orphan_workflow_runs(repo, owner_repo, dry_run, df_orphan_runs):
     dry_run       : dry run
     df_orphan_runs: pandas dataframe that contains orphan workflow runs
 
-    Return: total number of orphan workflow runs to be deleted
+    Return: total number of orphan workflow runs
     """
     console = Console()
     list_run_id = df_orphan_runs['run_id'].to_list()
@@ -236,29 +226,20 @@ def delete_orphan_workflow_runs(repo, owner_repo, dry_run, df_orphan_runs):
     if dry_run:
         console.print(f"\n([red]MOCK TO DELETE[/red]): [black]{list_run_id}[/black]\n")
     else:
-        # threads = []
         with Progress() as progress:
             overall_task = progress.add_task("[green]Processing data...\n", total=total_count)
 
-            # for workflow_run_id in list_run_id:
-            #     thread = threading.Thread(target=delete_workflow_runs, args=(total_count, repo, workflow_run_id))
-            #     threads.append(thread)
-            #     thread.start()
-            #     progress.update(overall_task, advance=1)
-
-            # for thread in threads:
-            #     thread.join()
-
             with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor:
                 for workflow_run_id in list_run_id:
                     futures = [executor.submit(delete_workflow_runs, total_count, repo, workflow_run_id)]
-                    for future in concurrent.futures.as_completed(futures):
+                    for _ in concurrent.futures.as_completed(futures):
                         progress.update(overall_task, advance=1)
 
     return len(list_run_id)
 
 
-def delete_active_workflow_runs_min_runs(repo, owner_repo, dry_run, min_runs, df):
+def delete_active_workflow_runs_min_runs(repo: Repository.Repository, owner_repo: str, dry_run: bool,
+                                         min_runs: int, df: pd.DataFrame) -> int:
     """
     Delete active workflow runs using min-runs option
 
@@ -286,7 +267,7 @@ def delete_active_workflow_runs_min_runs(repo, owner_repo, dry_run, min_runs, df
     """
     group_count_series = df_groupby_name.size()
     print('\n🐑 Active Workflow Runs (grouped by Workflow Name)')
-    print(f'{group_count_series}\n')
+    print(f'{group_count_series.to_string()}\n')
 
     """
     Filter groups (count > min_runs) into <class 'pandas.core.indexes.base.Index'>
@@ -303,8 +284,10 @@ def delete_active_workflow_runs_min_runs(repo, owner_repo, dry_run, min_runs, df
 
             """
             Get the row count for each group; Calculate the number of rows to remove; Get all the rows to remove
+            note: group_count_series.get() may result in int | None >> a tmp var in transit >> ensure group_count is int
             """
-            group_count_before_min_runs = group_count_series.get(filtered_group_names_index[i])
+            group_count_before_min_runs_tmp = group_count_series.get(filtered_group_names_index[i])
+            group_count_before_min_runs = group_count_before_min_runs_tmp if group_count_before_min_runs_tmp else 0
             group_count = group_count_before_min_runs - min_runs
             result_df_after_min_runs = result_df.head(group_count)
 
@@ -320,19 +303,9 @@ def delete_active_workflow_runs_min_runs(repo, owner_repo, dry_run, min_runs, df
                 console.print(f"([red]MOCK TO DELETE[/red]): "
                               f"[black]{result_df_after_min_runs['run_id'].to_list()}[/black]")
             else:
-                # threads = []
                 with Progress() as progress:
                     overall_task = progress.add_task("[green]Processing data...\n", total=group_count)
 
-                    # for index, row in result_df_after_min_runs.iterrows():
-                    #     thread = threading.Thread(target=delete_workflow_runs, args=(group_count, repo, row['run_id']))
-                    #     threads.append(thread)
-                    #     thread.start()
-                    #     progress.update(overall_task, advance=1)
-
-                    # for thread in threads:
-                    #     thread.join()
-
                     with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor:
                         for index, row in result_df_after_min_runs.iterrows():
                             futures = [executor.submit(delete_workflow_runs, group_count, repo, row['run_id'])]
@@ -346,7 +319,8 @@ def delete_active_workflow_runs_min_runs(repo, owner_repo, dry_run, min_runs, df
     return delete_active_workflow_runs_count
 
 
-def delete_active_workflow_runs_max_days(repo, owner_repo, dry_run, max_days, df):
+def delete_active_workflow_runs_max_days(repo: Repository.Repository, owner_repo: str, dry_run: bool,
+                                         max_days: int, df: pd.DataFrame) -> int:
     """
     Delete active workflow runs using max-days option
 
@@ -377,7 +351,7 @@ def delete_active_workflow_runs_max_days(repo, owner_repo, dry_run, max_days, df
     df_groupby_name = df.groupby('name')
     group_count_series = df_groupby_name.size()
     print('\n🐑 Active Workflow Runs (grouped by Workflow Name)')
-    print(f'{group_count_series}\n')
+    print(f'{group_count_series.to_string()}\n')
 
     """
     Filter workflow runs by created_at < cutoff_date with <class 'pandas.core.frame.DataFrame'>
@@ -413,27 +387,19 @@ def delete_active_workflow_runs_max_days(repo, owner_repo, dry_run, max_days, df
             """
             Get the row count for each group
             Accumulate the total number of rows to return for API estimate purpose
+            note: group_count_series.get() may result in int | None >> a tmp var in transit >> ensure group_count is int
             """
-            group_count = group_count_series.get(filtered_group_names_index[i])
+            group_count_tmp = group_count_series.get(filtered_group_names_index[i])
+            group_count = group_count_tmp if group_count_tmp else 0
             delete_active_workflow_runs_count += group_count
 
             print(f'\n🗑️ Deleting {group_count} workflow runs from {filtered_group_names_index[i]}')
             if dry_run:
                 console.print(f"([red]MOCK TO DELETE[/red]): [black]{result_df['run_id'].to_list()}[/black]")
             else:
-                # threads = []
                 with Progress() as progress:
                     overall_task = progress.add_task("[green]Processing data...\n", total=group_count)
 
-                    # for index, row in result_df.iterrows():
-                    #     thread = threading.Thread(target=delete_workflow_runs, args=(group_count, repo, row['run_id']))
-                    #     threads.append(thread)
-                    #     thread.start()
-                    #     progress.update(overall_task, advance=1)
-
-                    # for thread in threads:
-                    #     thread.join()
-
                     with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor:
                         for index, row in result_df.iterrows():
                             futures = [executor.submit(delete_workflow_runs, group_count, repo, row['run_id'])]
@@ -447,7 +413,7 @@ def delete_active_workflow_runs_max_days(repo, owner_repo, dry_run, max_days, df
     return delete_active_workflow_runs_count
 
 
-def delete_workflow_runs(count, repo, workflow_run_id):  # pragma: no cover
+def delete_workflow_runs(count: int, repo: Repository.Repository, workflow_run_id: int) -> int:
     """
     Delete workflow runs
 
@@ -463,11 +429,12 @@ def delete_workflow_runs(count, repo, workflow_run_id):  # pragma: no cover
         time.sleep(0.5)
         return workflow_run_id
 
-    except GithubException as e:
+    except GithubException as e:  # pragma: no cover
         print(f'❌ Failed to delete workflow run {workflow_run_id}: {e}')
+        return 0
 
 
-def get_api_estimate(orphan_runs_count, delete_runs_count):
+def get_api_estimate(orphan_runs_count: int, delete_runs_count: int) -> int:
     """
     Use dry-run to get API Usage Estimate
 
@@ -477,8 +444,8 @@ def get_api_estimate(orphan_runs_count, delete_runs_count):
 
     NOTE:
     1. this script consumes 3 API limit at the minimum
-    2. "delete workflow run" requires 2 API calls to 1) retrieve the workflow run object 2) call the delete method
-    3. every additional page (100 rows) on paginationlist adds an additional API call
+    2. every page (100 items) on paginationlist consumes an API call
+    3. "delete workflow run" requires 2 API calls to 1) retrieve the workflow run object 2) call the delete method
     """
     estimate = (
         ((orphan_runs_count + delete_runs_count) * 2) +
@@ -487,6 +454,39 @@ def get_api_estimate(orphan_runs_count, delete_runs_count):
     return estimate
 
 
+def write_data_dict(dry_run: bool, repo_url: str, min_runs: int, max_days: int, core_remaining: int,
+                    core_reset: datetime, core_usage_estimate: int, delete_active_workflow_runs_count: int,
+                    delete_orphan_workflow_runs_count: int):
+    """
+    Write data_dict to a file
+
+    Parameter(s):
+    dry_run                          : dry run
+    repo_url                         : repository url
+    max_days                         : maximum number of days to keep the run in a workflow
+    min_runs                         : minimum number of runs to keep in a workflow
+    core_remaining                   : core api rate limit remaining
+    core_reset                       : core api rate limit reset at
+    core_usage_estimate              : core api rate limit consumption for delete operation
+    delete_active_workflow_runs_count: number of active workflow runs to delete
+    delete_orphan_workflow_runs_count: number of orphan workflow runs to delete
+    """
+    data_dict = {}
+    data_dict.update({
+        "dry-run": dry_run,
+        "repo-url": repo_url,
+        "min-runs": min_runs,
+        "max-days": max_days,
+        "core-limit-remaining": core_remaining,
+        "core-limit-reset": str(core_reset),
+        "core-limit-usage-estimate": core_usage_estimate,
+        "delete-active-workflow-runs-count": delete_active_workflow_runs_count,
+        "delete-orphan-workflow-runs-count": delete_orphan_workflow_runs_count,
+    })
+    with open("data_dict.log", "w") as f:
+        json.dump(data_dict, f, indent=2)
+
+
 @click.command()
 @click.option("--dry-run", required=False, type=bool, default=True, show_default=True)
 @click.option("--repo-url", required=True, type=str, help="e.g. https://github.com/{owner}/{repo}")
@@ -495,31 +495,24 @@ def get_api_estimate(orphan_runs_count, delete_runs_count):
 @click.version_option(version=__version__)
 def main(dry_run, repo_url, min_runs, max_days):
     console = Console()
-    console.print(f"\n🚀 Starting to Delete GitHub Action workflows (dry-run: [red]{dry_run}[/red], "
-                  f"min-runs: [red]{min_runs}[/red], max-days: [red]{max_days}[/red])\n")
+    print(f"\n🚀 Starting to Delete GitHub Action workflows (dry-run: {dry_run}, " +
+          f"min-runs: {min_runs}, max-days: {max_days})\n")
 
-    """data to return"""
-    data_dict = {}
-    core_limit_usage_estimate = None
+    """initialize data"""
+    core_remaining = 0
+    core_reset = datetime.now() + timedelta(hours=1)
+    core_usage_estimate = 0
+    delete_orphan_workflow_runs_count = 0
+    delete_active_workflow_runs_count = 0
 
     try:
         gh = get_auth()
 
-        # """display initial core api rate limit info at the beginning"""
-        # core_limit = get_core_api_rate_limit(gh)
-        # core_limit_start = core_limit.used
-        # print('\n💥 Core API Rate Limit (start)')
-        # print(f'API rate limit          : {core_limit.limit}')
-        # print(f'API rate limit remaining: {core_limit.remaining}\n\n')
-
         """setup github repo object"""
         owner_repo = get_owner_repo(repo_url)
         repo = gh.get_repo(owner_repo)
 
         if check_user_inputs(repo, repo_url, min_runs, max_days):
-            delete_orphan_workflow_runs_count = 0
-            delete_active_workflow_runs_count = 0
-
             """
             get all workflow runs
             """
@@ -527,10 +520,8 @@ def main(dry_run, repo_url, min_runs, max_days):
             if (len(df_all_runs) > 0):
                 df_orphan_runs, df_active_runs, list_orphan_ids = break_down_df_all_runs(repo, df_all_runs)
             else:
-                # list_orphan_ids = []
                 df_active_runs = pd.DataFrame()
                 df_orphan_runs = pd.DataFrame()
-            # print(f'Number of orphan workflow IDs : {len(list_orphan_ids)}')
             print(f'\nTotal Number of workflow runs : {len(df_active_runs.index) + len(df_orphan_runs.index)}')
             print(f'Number of orphan workflow runs: {len(df_orphan_runs.index)}')
             print(f'Number of active workflow runs: {len(df_active_runs.index)}\n')
@@ -561,40 +552,24 @@ def main(dry_run, repo_url, min_runs, max_days):
             """
             display core api rate limit info and create a usage estimate
             """
-            core_limit = get_core_api_rate_limit(gh)
-            # core_limit_used = int(core_limit.used) - int(core_limit_start)
-            core_limit_reset = core_limit.reset
-            core_limit_remaining = core_limit.remaining
-            print('\n💥 Core API Rate Limit Info')
-            # print(f'API rate limit used     : {core_limit_used}')
-            print(f'API rate limit remaining: {core_limit_remaining}')
-            print(f'API rate limit reset at : {core_limit_reset} (UTC)\n')
-
+            core_remaining, core_reset = get_core_api_rate_limit(gh)
             if dry_run:
-                core_limit_usage_estimate =\
-                    get_api_estimate(delete_orphan_workflow_runs_count, delete_active_workflow_runs_count)
+                core_usage_estimate = get_api_estimate(delete_orphan_workflow_runs_count, delete_active_workflow_runs_count)
 
                 console.print('\n[blue]************************** API Usage Estimate ******************************[/blue]')
-                console.print(f'This delete can consume [red]{core_limit_usage_estimate}[/red] of your API limit.')
-                if (core_limit_remaining * 0.90) > core_limit_usage_estimate:
+                console.print(f'This delete can consume [red]{core_usage_estimate}[/red] units of your API limit.')
+                if (core_remaining * 0.90) > core_usage_estimate:
                     console.print('\nEnough API limit to run this delete now? ✅ yes')
                 else:
                     console.print('\nEnough API limit to run this delete now? ❌ no')
                     console.print('[red](segment this delete into multiple runs)[/red]')
                 console.print('[blue]****************************************************************************[/blue]')
 
-        data_dict.update({
-            "dry-run": dry_run,
-            "repo-url": repo_url,
-            "min-runs": min_runs,
-            "max-days": max_days,
-            "core-limit-remaining": core_limit_remaining,
-            "core-limit-reset": str(core_limit_reset),
-            "core-limit-usage-estimate": core_limit_usage_estimate,
-            "delete-active-workflow-runs-count": delete_active_workflow_runs_count,
-            "delete-orphan-workflow-runs-count": delete_orphan_workflow_runs_count,
-        })
-        return data_dict
+        """
+        write data_dict to a file for data feed to integrate with other tools
+        """
+        write_data_dict(dry_run, repo_url, min_runs, max_days, core_remaining, core_reset, core_usage_estimate,
+                        delete_active_workflow_runs_count, delete_orphan_workflow_runs_count)
 
     except Exception as e:
         print(f'❌ Exception Error: {e}')
@@ -602,5 +577,4 @@ def main(dry_run, repo_url, min_runs, max_days):
 
 
 if __name__ == '__main__':  # pragma: no cover
-    data_dict = main(standalone_mode=False)
-    print(f"::set-output name=result_dict::{data_dict}")
+    main()
